/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * DelaySourceFilter.cpp
 *
 *  Created on: 2011/02/24
 *      Author: HARK Development Team
 */

#include <map>
#include <deque>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Source.h>

#include <Logger.hpp>

using namespace std;
using namespace FD;
using namespace NSHarkBinaural;

class DelaySourceFilter;

DECLARE_NODE(DelaySourceFilter);
/*Node
 *
 * @name DelaySourceFilter
 * @category HARK:BINAURAL
 * @description Reduce short lifetime sources.
 *
 * @input_name INPUT
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @output_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @parameter_name DELAY
 * @parameter_type int
 * @parameter_value 100
 * @parameter_description Delay term (count). Delay term is (ADVANCE * DELAY) / SAMPLINGRATE.[default: 100].
 *
 END*/


namespace
{

/**
 * Sourceの発見時刻と生存期間を記憶する.
 */
struct SourceLifetime
{
    /// 発見時刻
    int discoveryTime;

    /// 生存期間
    int lifetime;

    SourceLifetime(int discoveryTime, int lifetime)
        : discoveryTime(discoveryTime), lifetime(lifetime)
    { }
};

/**
 * Sourceの発見時刻と生存期間のマップ.
 * キー: SourceのID
 * 値: Sourceの一番最初の発見時刻
 */
typedef map<int, SourceLifetime> LifetimeMap;

/**
 * 生存期間をカウントアップする.
 * もし既にtimeMapに存在するSourceなら、lifetimeをカウントアップさせ、
 * 新規のSourceならlifetimeを0として追加する。
 */
void CountupLifetime(LifetimeMap& timeMap, const Vector<ObjectRef>& sources, int currentTime)
{
    for (size_t i = 0; i < sources.size(); ++i)
    {
        RCPtr<Source> source = sources[i];
        LifetimeMap::iterator it = timeMap.find(source->id);
        if (it == timeMap.end())
        {
            timeMap.insert(LifetimeMap::value_type(source->id, SourceLifetime(currentTime, 0)));
        }
        else
        {
            it->second.lifetime++;
        }
    }
}

/**
 * 生存期間がスレッショルド以下のSourceを削除.
 */
void RemoveShortLifetimeSources(LifetimeMap& timeMap, int lifetimeThreshold, int currentTime)
{
    for (LifetimeMap::iterator it = timeMap.begin(); it != timeMap.end(); )
    {
        if (it->second.discoveryTime + lifetimeThreshold < currentTime && it->second.lifetime <= lifetimeThreshold)
        {
            LifetimeMap::iterator tmp = it;
            ++it;
            timeMap.erase(tmp);
        }
        else
        {
            ++it;
        }
    }
}

/**
 * 古いSourceを削除.
 */
void RemoveOldSources(LifetimeMap& timeMap, int lifetimeThreshold, int currentTime)
{
    for (LifetimeMap::iterator it = timeMap.begin(); it != timeMap.end(); )
    {
        if (it->second.discoveryTime + lifetimeThreshold + it->second.lifetime < currentTime)
        {
            LifetimeMap::iterator tmp = it;
            ++it;
            timeMap.erase(tmp);
        }
        else
        {
            ++it;
        }
    }
}

}


class DelaySourceFilter : public BufferedNode
{
    int inputID, outputID;
    int delay;
    LifetimeMap timeMap;

    deque<RCPtr<Vector<ObjectRef> > > sourceVectorQueue;

public:
    DelaySourceFilter(string nodeName, ParameterSet params) :
        BufferedNode(nodeName, params)
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        delay = getParameter<int>("DELAY");
        if (delay < 0) {
            delay = 0;
        }
        sourceVectorQueue.resize(delay);

        inputsCache[inputID].lookAhead = delay;

        inOrder = true;
    }

    virtual ~DelaySourceFilter()
    {
    }

    /**
     * ノードのパラメータ取得メソッド
     */
    template<typename T>
    T getParameter(const string& name)
    {
        return dereference_cast<T> (parameters.get(name));
    }

    void calculate(int output_id, int count, Buffer &out)
    {
        if (count == 0)
        {
            for (int i = 0; i < delay; ++i)
            {
                RCPtr<Vector<ObjectRef> > sourceInputPtr = getInput(this->inputID, count + i);
                Vector<ObjectRef>& sourceInput = *sourceInputPtr;
                CountupLifetime(timeMap, sourceInput, count + i);
            }
        }

        RCPtr<Vector<ObjectRef> > sourceInputPtr = getInput(this->inputID, count + delay);
        Vector<ObjectRef>& sourceInput = *sourceInputPtr;
        RCPtr<Source> source;

        //sourceVectorQueue.pop_front();
        //sourceVectorQueue.push_back(sourceInputPtr);

        CountupLifetime(timeMap, sourceInput, count + delay);
        RemoveShortLifetimeSources(timeMap, delay, count + delay);
        RemoveOldSources(timeMap, delay, count + delay);

        RCPtr<Vector<ObjectRef> > outputSourcesPtr(Vector<ObjectRef>::alloc(0));
        Vector<ObjectRef>& outputSources = *outputSourcesPtr;

        RCPtr<Vector<ObjectRef> > oldSourcesPtr = getInput(this->inputID, count);
        Vector<ObjectRef>& oldSources = *oldSourcesPtr;

        for (LifetimeMap::iterator it = timeMap.begin(); it != timeMap.end(); ++it)
        {
            if (it->second.lifetime >= delay)
            {
                //outputSources.push_back(it->second.source);
                /*
                for (size_t i = 0; i < sourceVectorQueue.front()->size(); ++i)
                {
                  source = sourceVectorQueue.front()->operator [](i);
                  if (source->id == it->first)
                  {
                    outputSources.push_back(source);
                    break;
                  }
                }
                */
                for (size_t i = 0; i < oldSources.size(); ++i)
                {
                    source = oldSources[i];
                    if (source->id == it->first)
                    {
                        outputSources.push_back(source);
                        break;
                    }
                }
            }
        }

        (*outputs[this->outputID].buffer)[count] = outputSourcesPtr;

    }
};
