/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */


/*
 * LocalizeBinaural.cc
 *
 *  Created on: 2010/12/09
 *      Author: HARK Development Team
 */

#include <iostream>
#include <string>
#include <typeinfo>
#include <complex>
#include <stdexcept>
#include <sstream>

#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>
#include <libharkio/libharkio.h>

#ifdef NDEBUG
#undef NDEBUG
#endif
#include <Debug.hpp>
#include <LocalizationSetting.hpp>
#include <InvalidInputException.hpp>
#include <Signal.hpp>
#include <Band.hpp>
#include <BGNEstimator.hpp>
#include <Peak.hpp>
#include <Logger.hpp>
#include <Cluster.hpp>
#include <Clustering.hpp>
#include <Harmonics.hpp>
#include <boolean.hpp>
#include <Localize.hpp>
#include <DirectionJudgeImpl.hpp>
#include <EventHistoryManager.hpp>
#include <IIPDS.hpp>
#include <HRTF.hpp>
#include <DirectionConverter.hpp>

#include <DimensionPromise/dimpro.hpp>

using namespace std;
using namespace boost;
using namespace FD;
using namespace NSHarkBinaural;
using namespace NSDimensionPromise;

class LocalizeBinaural;

DECLARE_NODE(LocalizeBinaural);
/*Node
 *
 * @name LocalizeBinaural
 * @category HARK:BINAURAL
 * @description Localize sounds
 *
 * @input_name AUDIO
 * @input_type Matrix<float>
 * @input_description Windowed sound data. A row index is channel, and a column index is frequency.
 *
 * @input_name SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency.
 *
 * @input_name BGNOISE
 * @input_type Matrix<float>
 * @input_description Background noise power spectrum. This is available only when you choose EXTERNAL method.
 *
 * @output_name SOURCES
 * @output_type Vector<ObjectRef>
 * @output_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @parameter_name BGNOISE_METHOD
 * @parameter_type string
 * @parameter_value EXTERNAL
 * @parameter_list EXTERNAL:INTERNAL
 * @parameter_description Method of background noise estimation. [default: EXTERNAL]
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz)[default: 16000].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length between adjacent frames (samples)[default: 160].
 *
 * @parameter_name PAUSE_LENGTH
 * @parameter_type int
 * @parameter_value 1000
 * @parameter_description Life duration of the source (ms)[default: 1000].
 *
 * @parameter_name MIN_SRC_INTERVAL
 * @parameter_type float
 * @parameter_value 20
 * @parameter_description The definition of "near" (Degree)[default: 20].
 *
 * @parameter_name PEAK_FMIN
 * @parameter_type float
 * @parameter_value 88
 * @parameter_description Minimum frequency of peak extraction (Hz)[default: 88].
 *
 * @parameter_name PEAK_FMAX
 * @parameter_type float
 * @parameter_value 3000
 * @parameter_description Maximum frequency of peak extraction (Hz)[default: 3000].
 *
 * @parameter_name PEAK_THRESHOLD
 * @parameter_type float
 * @parameter_value 15
 * @parameter_description Threshold to regard a power spectrum as a peak (db)[default: 15].
 *
 * @parameter_name LOCALIZE_CHANNEL
 * @parameter_type string
 * @parameter_value LEFT
 * @parameter_list LEFT:RIGHT
 * @parameter_description Channel to use for localization [default: LEFT].
 *
 * @parameter_name HARMONIC_THRESHOLD
 * @parameter_type int
 * @parameter_value 2
 * @parameter_description Threshold to regard a sound as a human harmonics [default: 2]. \nThe less this value is, the easier a sound is regarded as a human harmonics.
 *
 * @parameter_name EVENT_RATIO_THRESHOLD
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Event update ratio threshold [default: 0].
 *
 * @parameter_name CLUSTERING_THRESHOLD
 * @parameter_type float
 * @parameter_value 5
 * @parameter_description Threshold to regard two clusters as two deferent clusters [default: 5].
 *
 * @parameter_name LOCALIZE_METHOD
 * @parameter_type string
 * @parameter_value EPIPOLAR
 * @parameter_list EPIPOLAR:HRTFMESURED
 * @parameter_description Method of localization [default: EPIPOLAR].
 *
 * @parameter_name DIRPROB_THRESHOLD
 * @parameter_type float
 * @parameter_value 0.95
 * @parameter_description Threshold of probability [default: 0.95].
 *
 * @parameter_name ROUND_COUNT
 * @parameter_type int
 * @parameter_value 3
 * @parameter_description Search range for local maximum [default: 3].
 *
 * @parameter_name HARK_DAT
 * @parameter_type string
 * @parameter_value hark.dat
 * @parameter_description Transfer function file.
 *
 * @parameter_name HARK_DAT_CHANNELS
 * @parameter_type object
 * @parameter_value <Vector<int> 0 1>
 * @parameter_description Channels of HARK_DAT. The first element is L channel, the second element is R channel.
 *
 END*/


string HarmonicSoundToString(const HarmonicSound& sound)
{
    ostringstream oss;
    oss << format("ID %d") % (int)sound.eventID;
    if (sound.bands->size() > 0)
    {
        oss << format(" %4d Hz hnum %d dir %3d prob %2d [") %
            static_cast<int>(sound.bands->at(0).band.Freq().X()) %
            sound.bands->size() %
            gDirConv.ToDegree(sound.directions->at(0).first) %
            rint(sound.directions->at(0).second.Get() * 100);
        for (size_t i = 0, j = 0; i < sound.bands->size() && j < sound.bands->back().ratio; ++j)
        {
            if (sound.bands->at(i).ratio == j + 1)
            {
                oss << format(" %3d") % sound.bands->at(i).band.Index();
                i++;
            }
            else
            {
                oss << format(" %3d") % 0;
            }
        }
        oss << "]";
    }
    return oss.str();
}

class LocalizeBinaural: public BufferedNode
{
    int audioInputID, spectrumInputID, bgnoiseInputID;
    int sourcesOutputID;
    int samplingRate, advance, pauseLength;
    bool useInternalBGNoise;

    Frequency peakFMin, peakFMax;
    Channel::Kind localizeChannel;
    int harmonicThreshold;
    float peakThreshold;
    float eventUpdateRatioThreshold;
    float clusteringThreshold;
    float dirprobThreshold;
    int roundCount;

    LocalizationSetting::ptr_t localizationSetting; // 定位設定

    BGNEstimator bgnEstimator; // バックグラウンド計算機
    Spectrum::ptr_t internalBGNoise; // 内部的に計算したバックグラウンドノイズ
    EventHistoryManager ehman;

    RCPtr<Vector<ObjectRef> > lastSources;

    shared_ptr<IIPDS> iipds;
    String localizationMethod;

    Logger& logger;

    vector<int> knownEvent; ///< 既に出力済みのイベント
public:
    LocalizeBinaural(string nodeName, ParameterSet params) :
        BufferedNode(nodeName, params),
        bgnEstimator(10),
        ehman(),
        lastSources(new Vector<ObjectRef>()),
        logger(Logger::GetLogger(Logger::GLOBAL_LOGGER_NAME))
    {

        /*
         * 入出力端子の設定
         */
        audioInputID = addInput("AUDIO");
        spectrumInputID = addInput("SPECTRUM");
        bgnoiseInputID = addInput("BGNOISE");
        sourcesOutputID = addOutput("SOURCES");

        samplingRate = getParameter<int>("SAMPLING_RATE");
        advance = getParameter<int>("ADVANCE");
        String bgnoiseMethod = object_cast<String>(parameters.get("BGNOISE_METHOD"));
        if (bgnoiseMethod == "INTERNAL")
        {
            useInternalBGNoise = true;
        }
        else
        {
            useInternalBGNoise = false;
        }

        pauseLength = getParameter<int>("PAUSE_LENGTH");
        float minSrcInterval = getParameter<float>("MIN_SRC_INTERVAL");

        timeval validPausePeriod = { pauseLength / 1000, (pauseLength * 1000) % 10000000 };
        ehman.SetValidPausePeriod(validPausePeriod);
        ehman.SetNearDirectionThreshold(Angle::FromDegree(minSrcInterval / 2));

        peakFMin = Frequency::FromHz(getParameter<float>("PEAK_FMIN"));
        peakFMax = Frequency::FromHz(getParameter<float>("PEAK_FMAX"));

        String& localizeChannelString = object_cast<String>(parameters.get("LOCALIZE_CHANNEL"));
        localizeChannel = localizeChannelString == "LEFT" ? Channel::L : Channel::R;

        harmonicThreshold = getParameter<int>("HARMONIC_THRESHOLD");

        peakThreshold = getParameter<float>("PEAK_THRESHOLD");

        eventUpdateRatioThreshold = getParameter<float>("EVENT_RATIO_THRESHOLD");

        clusteringThreshold = getParameter<float>("CLUSTERING_THRESHOLD");

        ehman.SetEventUpdateRatioThreshold(eventUpdateRatioThreshold);

        localizationMethod = object_cast<String>(parameters.get("LOCALIZE_METHOD"));

        dirprobThreshold = getParameter<float>("DIRPROB_THRESHOLD");

        roundCount = getParameter<int>("ROUND_COUNT");

        inOrder = true;

        if (localizationMethod == "EPIPOLAR")
        {
            // インデックス間隔5度（インデックスが増えるほど右）、90度（左）がインデックス0
            gDirConv = DirectionConverter(-5, 90);
        }
        else if (localizationMethod == "HRTFMESURED")
        {
            // 仮の値
            gDirConv = DirectionConverter(0, 90);
        }
    }

    virtual ~LocalizeBinaural()
    {
    }

    /**
     * ノードのパラメータ取得メソッド
     */
    template<typename T>
    T getParameter(const string& name)
    {
        return dereference_cast<T> (parameters.get(name));
    }

    void count0Initialize(int output_id, Buffer &out)
    {
        try
        {
            // 入力を取得する
            RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(this->spectrumInputID, 0);
            Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;

            // 音源定位の設定
            this->localizationSetting = LocalizationSetting::ptr_t(new LocalizationSetting());
            this->localizationSetting->ndft = (spectrumInput.ncols() - 1) * 2;
            this->localizationSetting->samplingFrequency = Frequency::FromHz(this->samplingRate);

            if (this->useInternalBGNoise)
            {
                this->internalBGNoise = Spectrum::ptr_t(new Spectrum(this->localizationSetting->ndft));
            }

            if (localizationMethod == "HRTFMESURED")
            {
                harkio::HARK hark;
                //hark.readFile("/home/harkuser/myaudition/hark.dat");
                hark.readFile(object_cast<String>(parameters.get("HARK_DAT")));

                Vector<int> channels = object_cast<Vector<int> >(parameters.get("HARK_DAT_CHANNELS"));
                if (channels.size() != 2)
                {
                    throw new NodeException(this, "The size of HARK_DAT_CHANNELS != 2", __FILE__, __LINE__);
                }

                iipds = HRTFToIIPDS(hark, channels[0], channels[1]);
                LogInfo(logger, format("iipds: fs = %d, nfft=%d, nfreq=%d, dirtick=%d, ndir=%d, begindeg=%d")
                    % iipds->fs % iipds->nfft % iipds->nfreq % iipds->dirtick % iipds->ndir % iipds->begindeg);

                gDirConv = DirectionConverter(iipds->dirtick, iipds->begindeg);
            }
        }
        catch (const std::exception& e)
        {
            // std::exceptionをNodeExceptionに変換する
            throw new NodeException(this, e.what(), __FILE__, __LINE__);
        }

#if 0
        ofstream ipdFile("ipd.out");
        for (size_t dir = 0; dir < iipds->ipd.size(); ++dir)
        {
            for (size_t i = 0; i < iipds->ipd[dir].m.size(); ++i)
            {
                ipdFile << iipds->ipd[dir].m[i] << '\n';
            }
        }

        ofstream iidFile("iid.out");
        for (size_t dir = 0; dir < iipds->iid.size(); ++dir)
        {
            for (size_t i = 0; i < iipds->iid[dir].m.size(); ++i)
            {
                iidFile << iipds->iid[dir].m[i] << '\n';
            }
        }
#endif
    }

    void calculate(int output_id, int count, Buffer &out)
    {

        LogFine(logger, "inputting variables");

        // 入力を取得する
        RCPtr<Matrix<float> > audioInputPtr = getInput(this->audioInputID, count);
        Matrix<float>& audioInput = *audioInputPtr;
        RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(this->spectrumInputID, count);
        Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;

        // 出力変数
        RCPtr<Vector<ObjectRef> > sourcesPtr(Vector<ObjectRef>::alloc(0));
        Vector<ObjectRef>& sources = *sourcesPtr;

        LogFine(logger, "inspecting variables");

        ostringstream logOss;

        // 入力を検査する
        if (audioInput.nrows() != 2 || spectrumInput.nrows() != 2)
        {
            throw new InvalidInputException(
                (format("Input matrix is not 2ch, but %dch") % audioInput.nrows()).str());
        }

        if (count == 0)
        {
            // 最初のカウントのみ初期化する
            count0Initialize(output_id, out);
        }

        // 現在時刻(秒)
        Time currentTime = static_cast<double>(count) * static_cast<double>(advance) / this->localizationSetting->samplingFrequency;

        // スペクトルを取得する
        shared_ptr<vector<Spectrum::ptr_t> > spectrumArray = ToSignalArray<Spectrum::ElementType>(spectrumInput);
        LRSpectra::ptr_t lrspectra(new LRSpectra(spectrumArray->at(0), spectrumArray->at(1)));
        if (lrspectra.get() == 0)
        {
            logger.Warning("unable to create lrspectra");
        }
        else
        {
            logger.Fine("lrspectra created");
        }

        // スペクトルを正規化
        if (IsEqualSignalSize(lrspectra))
        {
            *lrspectra->Left() /= this->localizationSetting->ndft;
            *lrspectra->Right() /= this->localizationSetting->ndft;
            if (count < 3)
            {
                /*
                cout << "spec data 0ch: ";
                for (int i = 0; i < 10; ++i)
                {
                  cout << lrspectra->Left()->At(i) << " ";
                }
                cout << endl;
                */
            }
        }

        // ルームノイズを取得する
        logger.Fine("getting room noise");
        LRUnspectra::ptr_t lrbgnoise;
        if (this->useInternalBGNoise)
        {
            // 内蔵の推定器でBGNを推定する
            logger.Fine("internal bgn estimating");
            this->bgnEstimator.Estimate(lrspectra);
            if (this->bgnEstimator.Estimated())
            {
                logger.Fine("background noise estimated");
                lrbgnoise = this->bgnEstimator.BGNoise();
            }
        }
        else
        {
            // 外部の推定器で推定したBGNを使う
            ObjectRef bgnoiseInputPtr0 = getInput(this->bgnoiseInputID, count);
            if (!bgnoiseInputPtr0.isNil())
            {
                logger.Fine("background noise got (external)");
                RCPtr<Matrix<float> > bgnoiseInputPtr = bgnoiseInputPtr0;
                Matrix<float>& bgnoiseInput = *bgnoiseInputPtr;
                shared_ptr<vector<Unspectrum::ptr_t> > bgnoiseArray = ToSignalArray<Unspectrum::ElementType>(bgnoiseInput);
                lrbgnoise = LRUnspectra::ptr_t(new LRUnspectra(bgnoiseArray->at(0), bgnoiseArray->at(1)));
            }
        }

        // フーリエ変換の分割数とサンプリング周波数の対応管理
        BandInformation bi(this->localizationSetting->samplingFrequency, this->localizationSetting->ndft);

        LRUnspectra::ptr_t lrpower = LRSignalFactory<Unspectrum::ElementType>::Create(lrspectra->Left()->Size());
        for (size_t i = 0; i < lrspectra->Left()->Size(); ++i)
        {
            lrpower->Left()->At(i) = Decibel(lrspectra->Left()->At(i));
            lrpower->Right()->At(i) = Decibel(lrspectra->Right()->At(i));
        }

        if (lrbgnoise.get() != 0)
        {
            Peak::LRPeaksPtr peak = Peak::GetPeaks(lrbgnoise, lrpower,
                                                   bi.CreateBand(this->peakFMin), bi.CreateBand(this->peakFMax), bi, this->peakThreshold);

            Peak::PeakVectorPtr peak_ = this->localizeChannel == Channel::L ? peak->Left() : peak->Right();
            if (peak_->size() > 0)
            {
                logOss.str("");
                logOss << "got peak (size = " << peak_->size() << "): ";
                for (size_t i = 0; i < peak_->size(); ++i)
                {
                    logOss << "(" << peak_->at(i).band.Index() << ':' << peak_->at(i).power << ") ";
                }
                logOss << endl;
                LogFinest(logger, logOss.str());

                /*
                ofstream ofs("peak.csv", ios::out | ios::app);
                for (size_t i = 0, j = 0; i < this->localizationSetting->ndft / 2 + 1 && j < peak_->size(); ++i)
                {
                  if (peak_->at(j).band.Index() == i)
                  {
                    ofs << peak_->at(j).power << '\n';
                    j++;
                  }
                  else
                  {
                    ofs << 0 << '\n';
                  }
                }
                ofs << endl;
                */
            }

            shared_ptr<vector<shared_ptr<Cluster> > > clusters =
                Clustering::Execute(this->localizeChannel == Channel::L ? peak->Left() : peak->Right(), clusteringThreshold);

            if (clusters.get() == 0)
            {
                LogSevere(logger, "clusters is null");
            }

            if (clusters->size() > 0)
            {
                logOss.str("");
                logOss << "got cluster (size = " << clusters->size() << "): ";
                for (size_t i = 0; i < clusters->size(); ++i)
                {
                    logOss << "band" << clusters->at(i)->BandAndPowerList()[0].band.Index() << ':' << clusters->at(i)->BandAndPowerList().size() << ' ';
                }
                logOss << endl;
                LogFinest(logger, logOss.str());
            }

            shared_ptr<vector<Band> > proxy =
                Clustering::GetProxy(clusters, this->localizeChannel == Channel::L ? lrbgnoise->Left() : lrbgnoise->Right());

            /*
            if (this->localizeChannel == Channel::L)
            {
              proxy = Clustering::GetProxy(clusters, lrbgnoise->Left());
            }
            else
            {
              proxy = Clustering::GetProxy(clusters, lrbgnoise->Right());
            }
            */
            if (proxy.get() == 0)
            {
                LogSevere(logger, "proxy is null");
            }

            if (proxy->size() > 0)
            {
                logOss.str("");
                logOss << format("got proxy (size = %d):") % proxy->size();
                for (size_t i = 0; i < proxy->size(); ++i)
                {
                    logOss << format(" %d:%d") % i % proxy->at(i).Index();
                }
                logOss << endl;
                LogFinest(logger, logOss.str());
            }

            Harmonics::HarmonicSoundVectorPtr sounds = Harmonics::ExtractAllHarmonics(proxy);
            if (sounds->size() > 0)
            {
                logOss.str("");
                logOss << format("got all harmonics (size = %d):") % sounds->size();
                for (size_t i = 0; i < sounds->size(); ++i)
                {
                    logOss << format(" %d: size=%d,fund=%d") % i % sounds->at(i)->bands->size() % sounds->at(i)->bands->at(0).band.Index();
                }
                logOss << endl;
                LogFinest(logger, logOss.str());
            }

            sounds = Harmonics::GetHumanHarmonics(sounds, this->localizationSetting->ndft / 2 + 1, bi, this->harmonicThreshold);
            if (sounds->size() > 0)
            {
                logOss.str("");
                logOss << format("got human harmonics (size = %d):") % sounds->size();
                for (size_t i = 0; i < sounds->size(); ++i)
                {
                    logOss << format(" %d: size=%d,fund=%d") % i % sounds->at(i)->bands->size() % sounds->at(i)->bands->at(0).band.Index();
                }
                logOss << endl;
                LogFinest(logger, logOss.str());
            }

            if (sounds->size() > 0)
            {
                IPD::ptr_t ipd(new IPD(lrspectra, Channel::L));
                IID::ptr_t iid(new IID(lrspectra, Channel::L));

                shared_ptr<DirectionJudge> dirJudge;

                if (localizationMethod == "HRTFMESURED")
                {
                    dirJudge = shared_ptr<DirectionJudge>(new JudgeHRTFMeasured(iipds));
                }
                else
                {
                    dirJudge = shared_ptr<DirectionJudge>(new JudgeEpipolar(Length::FromMeter(1.0), Length::FromMillimeter(90)));
                }

#if 1
                // 音源定位
                Localize::DirectionProbabilityVectorVectorPtr directions =
                    Localize::GetDirections(sounds, bi,
                                            *dirJudge
                                            , ipd, iid, MethodFlag::BOTH, true);

                // 定位された方向をロギング
                logOss.str("");
                for (size_t i = 0; i < directions->size(); ++i)
                {
                    if (directions->at(i)->size() > 0)
                    {
                        logOss << "got directions of sound[" << i << "]: " << directions->at(i)->size()
                               << ", highest prob direction is " << gDirConv.ToDegree(directions->at(i)->at(0).first)
                               << " (" << directions->at(i)->at(0).second.Get() * 100 << "%)" << endl;
                    }
                    sounds->at(i)->directions = directions->at(i);
                }
                logOss << endl;
                LogFinest(logger, logOss.str());

                // イベントの更新
                timeval currentTimeval;
                currentTimeval.tv_sec = static_cast<time_t>(floor(currentTime.ToSec()));
                currentTimeval.tv_usec = static_cast<suseconds_t>(floor((currentTime.ToSec() - currentTimeval.tv_sec) * 1000000));
                ehman.CheckSoundHistory(sounds, currentTimeval);

                logOss.str("");
                for (size_t i = 0; i < sounds->size(); ++i)
                {
                    logOss << format("%lf sec") % (currentTime.ToSec());
                    logOss << format(" No %d : ") % (i + 1) << HarmonicSoundToString(*sounds->at(i)) << endl;
                }
                LogFinest(logger, logOss.str());

            }

            // すべてのイベントをSource構造体へ変換
            for (list<EventHistory>::const_iterator it = ehman.GetAllEvents().begin(); it != ehman.GetAllEvents().end(); ++it)
            {
                if (it->Status() == EventStatus::DURATION || it->Status() == EventStatus::PAUSE)
                {
                    RCPtr<Source> src(new Source());
                    src->id = it->ID();
                    src->x[0] = cos(it->Direction().ToRadian());
                    src->x[1] = sin(it->Direction().ToRadian());
                    sources.push_back(src);
                }
            }

#else
                // 音源定位
                Localize::DirectionProbabilityVectorVectorPtr directions =
                    Localize::GetDirections(sounds, bi,
                                            *dirJudge
                                            , ipd, iid, MethodFlag::BOTH, false);

                // 角度ごとに確率を足す
                vector<double> prob(directions->at(0)->size());
                for (size_t i = 0; i < directions->size(); ++i)
                {
                    Localize::DirectionProbabilityVector& direction = *directions->at(i);
                    for (size_t dir = 0; dir < prob.size(); ++dir)
                    {
                        prob[dir] += direction.at(dir).second;
                    }
                }

                // 正規化
                for (size_t dir = 0; dir < prob.size(); ++dir)
                {
                    prob[dir] /= directions->size();
                }

                stringstream ss;
                string countString;
                ss << count;
                ss >> countString;
                ofstream dirFile((countString + ".dir").c_str());
                for (size_t dir = 0; dir < prob.size(); ++dir)
                {
                    dirFile << prob[dir] << endl;
                }

                // ピークを探す
                vector<size_t> dirPeaks;

                double probThres = dirprobThreshold;
                for (size_t dir = 0; dir < prob.size(); ++dir)
                {
                    if (prob[dir] > probThres)
                    {
                        bool isPeak = true;
                        for (size_t round = dir < roundCount ? 0 : dir - roundCount;
                                round < (dir + roundCount < prob.size() ? dir + roundCount : prob.size());
                                round++)
                        {
                            if (dir != round)
                            {
                                if (prob[dir] <= prob[round])
                                {
                                    isPeak = false;
                                    break;
                                }
                                else if (round < dir && prob[round] >= prob[round + 1])
                                {
                                    isPeak = false;
                                    break;
                                }
                                else if (dir < round && prob[round - 1] <= prob[round])
                                {
                                    isPeak = false;
                                    break;
                                }
                            }
                        }
                        if (isPeak)
                        {
                            dirPeaks.push_back(dir);
                        }
                    }
                }

                for (vector<size_t>::iterator it = dirPeaks.begin(); it != dirPeaks.end(); ++it)
                {
                    Angle dir = Angle::FromDegree(static_cast<signed long>(*it) * TICKDEGREE - 90);
                    RCPtr<Source> src(new Source());
                    src->x[0] = cos(dir.ToRadian());
                    src->x[1] = sin(dir.ToRadian());
                    sources.push_back(src);
                }

            }
#endif
        }

        for (Vector<ObjectRef>::iterator it = sourcesPtr->begin(); it != sourcesPtr->end(); ++it)
        {
            RCPtr<Source> src = *it;
            if (find(knownEvent.begin(), knownEvent.end(), src->id) == knownEvent.end())
            {
                cout << format("SOURCE ID %d DIR %f") % src->id % (atan2(src->x[1], src->x[0]) * 180 / M_PI) << endl;
                knownEvent.push_back(src->id);
            }
        }

        lastSources = sourcesPtr;
        (*outputs[this->sourcesOutputID].buffer)[count] = sourcesPtr;
    }

};

